"{ Encoding: utf8 }"

"
 COPYRIGHT (c) 1999/2015 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
"{ Package: 'stx:libwidg2' }"

"{ NameSpace: Smalltalk }"

Object subclass:#AbstractHierarchicalItem
	instanceVariableNames:'parent children'
	classVariableNames:''
	poolDictionaries:''
	category:'Views-Support'
!

!AbstractHierarchicalItem class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1999/2015 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    Hierarchical Items are mostly like Models, but the list of
    dependencies are kept by its HierarchicalList.
    The class is used to build up hierarchical trees.

    2015 update:
        the original HierarchicalItem has been refactored into this abstract class,
        which provides all the mechanisms, but leaves the concrete representation
        of some slots open.
        These are:
            - if and how the geometry information (width + height) are cached,
            - if and how the expanded-state is remembered.
            - if and how the underlying model is fetched

        The old class used private slots for the first three (width-height-isExpanded),
        but did not keep a reference to the model. This leads to a very poor performance,
        as many algorithms degenerated to O(n log(n)) or even O(n^2) time behavior,
        as the model was fetched by walking along the parent chain - sometimes for every item
        in a long list.

        The old class is still around and may be used for small trees,
        but we recommend rewriting applications to use the new CompactHierarchicalItem
        class, which behaves the same on the outside, but uses clever tricks to be both more
        space efficient (saving 2 slots) and time efficient (caching the model).
        For this, make sure that the subclass does not access the instvars isExpanded, width and height
        directly, but uses the getters isExpanded, width and height and the setters setExpanded:, width: and height:

    [Instance variables:]
        parent      <Item, List or nil>         parent or my HierarchicalList.
        children    <Collection or nil>         list of children

    [author:]
        Claus Atzkern
        Claus Gittinger (redesign and refactoring)

    [see also:]
        HierarchicalItem (the old item class)
        HierarchicalList (typical model)
        HierarchicalListView (typical user of me)
"
! !

!AbstractHierarchicalItem class methodsFor:'instance creation'!

new
    ^ (self basicNew) initialize
!

parent:aParent
    |item|

    item := self new.
    item parent:aParent.
  ^ item
! !

!AbstractHierarchicalItem class methodsFor:'protocol'!

doResetExtentOnChange
    "true: the extent of the item is reset if a change
     notification is raised from the item. the default is true
    "
    ^ true
! !

!AbstractHierarchicalItem class methodsFor:'queries'!

isAbstract
    "Return if this class is an abstract class.
     True is returned here for myself only; false for subclasses.
     Abstract subclasses must redefine this again."

    ^ self == AbstractHierarchicalItem.
! !

!AbstractHierarchicalItem methodsFor:'accessing'!

getChildren
    "returns the children as they are present (or not); not going to the model..."

    ^ children
!

level
    "returns the level starting with 0 for the root"

    |item level|

    item := self.
    level := 0.

    [ (item := item parentOrModel) notNil] whileTrue:[
        level := level + 1.
        level > 10000 ifTrue:[
            self halt:'possibly recursive item hierarchy'.
            ^ self.
        ].
    ].
    ^ level

    "Modified: / 09-07-2010 / 08:56:27 / cg"
!

parent
    "returns the parent or nil"

    |p|

    p := parent.
    ^ (p notNil and:[p isHierarchicalItem]) 
        ifTrue:[p]
        ifFalse:[nil]

    "Modified: / 01-08-2018 / 10:18:37 / Claus Gittinger"
    "Modified: / 01-08-2018 / 16:15:07 / Stefan Vogel"
!

parent:aParent
    "set the parent (or the model if the item is the root item)"

    parent := aParent
!

rootItem
    "returns the root item"

    parent isHierarchicalItem ifTrue:[
        ^ parent rootItem
    ].
    ^ self
! !

!AbstractHierarchicalItem methodsFor:'accessing-children'!

at:anIndex
    "return the child at anIndex if valid;
     if the index is invalid, nil is returned"

    ^ self at:anIndex ifAbsent:nil
!

at:anIndex ifAbsent:exceptionBlock
    "return the child at anIndex if valid; if the index is
     invalid, the result of evaluating the exceptionBlock is returned."

    |list|

    (list := self children) notNil ifTrue:[
        ^ list at:anIndex ifAbsent:exceptionBlock
    ].
    ^ exceptionBlock value
!

at:anIndex put:anItem
    "replace a child by a new item. return anItem (sigh)"

    |children oldItem visIndex model expFlag|

    anItem isNil ifTrue:[
        self removeFromIndex:anIndex toIndex:anIndex.
        ^ nil
    ].
    anItem parent:self.

    (model := self model) isNil ifTrue:[
        self children at:anIndex put:anItem.
        ^ anItem
    ].

    model synchronized:[
        children := self children.
        oldItem  := children at:anIndex.

        oldItem isExpanded ifTrue:[
            oldItem collapse
        ].
        visIndex := model identityIndexOf:oldItem.
        expFlag  := anItem isExpanded.
        anItem setExpanded:false.

        children at:anIndex put:anItem.

        visIndex ~~ 0 ifTrue:[
            model at:visIndex put:anItem.
        ].
        self changed:#redraw.
        expFlag ifTrue:[ anItem expand ].
    ].
    ^ anItem

    "Modified: / 28-07-2018 / 15:51:06 / Claus Gittinger"
!

children:aListOfChildren
    "set a new list of children"

    self synchronized:[
        self removeAll.
        self addAll:aListOfChildren beforeIndex:1
    ].
    ^ aListOfChildren

    "Modified: / 28-07-2018 / 15:47:25 / Claus Gittinger"
!

first
    "returns the first child
    "
    ^ self at:1
!

last
    "returns the last child"

    ^ self at:(self size)
!

second
    "returns the second child"

    ^ self at:2
! !

!AbstractHierarchicalItem methodsFor:'accessing-hierarchy'!

collapse
    "hide all my subitems"

    |visChd index|

    self canCollapse ifTrue:[
        self setExpanded:false.

        self synchronized:[
            (index := self listIndex) notNil ifTrue:[
                "/ do not call :#size: children will be autoloaded !!!!
                (visChd := children size) ~~ 0 ifTrue:[
                    self nonCriticalFrom:1 to:nil do:[:el|
                        visChd := visChd + el numberOfVisibleChildren
                    ].
                    self model itemRemoveFromIndex:(index + 1) toIndex:(index + visChd).
                ].
                index ~~ 0 ifTrue:[self hierarchyChanged].
            ]
        ]
    ]

    "Modified: / 28-07-2018 / 15:47:29 / Claus Gittinger"
!

enforcedExpand
    "expand children - even if there are no children,
     the item is initially expanded (but this might be undone later,
     when we know that no children are there"

    self expand:true
!

expand
    "expand children - but only if there are children 
     (i.e. this cannot be used before the childInfo is valid;
      aka not before the updateTask came along this item)"

    self expand:false
!

expand:enforced
    "expand children"

    "/ test whether the item already is expanded; #canExpand could be redefined
    "/ without checking whether the node is expanded (happens already) !!

    self isExpanded ifTrue:[ ^ self ].
    (enforced not and:[self canExpand not]) ifTrue:[ ^ self ].

    self synchronized:[
        |index list|

        (index := self listIndex) notNil ifTrue:[
            "/ must set expand-flag to false, otherwise change notifications
            "/ are raised during lazy auto creation (to the list).
            self setExpanded:false.
            list := self children.
            list notEmptyOrNil ifTrue:[
                self setExpanded:true. 
                list := OrderedCollection new.
                self addVisibleChildrenTo:list.
                self model itemAddAll:list beforeIndex:(index + 1).
            ].
            index ~~ 0 ifTrue:[self hierarchyChanged].
        ] ifFalse:[
            self setExpanded:true
        ]
    ].

    "Modified: / 08-03-2018 / 21:06:13 / stefan"
    "Modified: / 28-07-2018 / 15:47:47 / Claus Gittinger"
!

expandLevels:numLevels
    "expand children numLevels down"

    numLevels == 0 ifTrue:[^ self].
    self children do:[:each |
        each expand.
        each expandLevels:(numLevels - 1).
    ].
!

expandLevels:numLevels max:maxNumExpandedHolder
    "expand children numLevels down"

    numLevels == 0 ifTrue:[^ self].
    maxNumExpandedHolder value <= 0 ifTrue:[^ self].

    self children do:[:each |
        each expand.
    ].

    maxNumExpandedHolder value:(maxNumExpandedHolder value - self children size).
    self children do:[:each |
        each expandLevels:(numLevels - 1) max:maxNumExpandedHolder.
    ].
!

labelPath
    "return my label-path as an ordered collection of individual labels"

    ^ ({self} , self allParents) reverse collect:#label

    "Created: / 16-10-2017 / 13:09:10 / cg"
    "Modified: / 16-10-2017 / 14:16:46 / cg"
!

makeVisible
    "expand all my parents"

    (parent notNil and:[parent isHierarchicalItem]) ifTrue:[
        self synchronized:[
            parent makeVisible.
            parent expand.
        ]
    ].

    "Modified: / 28-07-2018 / 15:47:52 / Claus Gittinger"
!

recursiveCollapse
    "collapse all item and sub items
     **** must be expanded
    "
    self synchronized:[
        |visChd index|

        self canCollapse ifTrue:[
            (index := self listIndex) notNil ifTrue:[
                "/ do not call :#size: children will be autoloaded !!!!
                (visChd := children size) ~~ 0 ifTrue:[
                    self nonCriticalFrom:1 to:nil do:[:el|
                        visChd := visChd + el numberOfVisibleChildren
                    ].
                ].
                self recursiveSetCollapsed.

                visChd ~~ 0 ifTrue:[
                    self model itemRemoveFromIndex:(index + 1)
                                           toIndex:(index + visChd)
                ].
                index ~~ 0 ifTrue:[
                    self hierarchyChanged
                ]
            ] ifFalse:[
                self recursiveSetCollapsed
            ]
        ]
    ]

    "Modified (format): / 28-07-2018 / 15:48:20 / Claus Gittinger"
!

recursiveExpand
    "expand children and sub-children
     Precondition: must be collapsed"

    |index list|

    "/ test whether the item already is expanded; #canExpand could be redefined
    "/ without checking whether the node is expanded (happens actually) !!
    self isExpanded ifTrue:[ ^ self ].
    self canExpand ifFalse:[ ^ self ].

    self setExpanded:true.

    self synchronized:[
        self size ~~ 0 ifTrue:[
            index := self listIndex.    "/ get the visible list index

            index isNil ifTrue:[        "/ not visible
                self nonCriticalFrom:1 to:nil do:[:el|
                    el setExpanded:true
                ].
            ] ifFalse:[
                list := OrderedCollection new.
                self recursiveSetExpandedAndAddToList:list.
                self model itemAddAll:list beforeIndex:(index + 1).

                index ~~ 0 ifTrue:[self hierarchyChanged]
            ]
        ]
    ].

    "Modified (format): / 28-07-2018 / 13:56:15 / Claus Gittinger"
    "Modified (comment): / 02-08-2018 / 16:16:17 / Stefan Vogel"
!

recursiveToggleExpand
    "if the item is collapsed, the item and all its sub-items
     are expanded otherwise collapsed"

    self isExpanded ifTrue:[
        self recursiveCollapse
    ] ifFalse:[
        self recursiveExpand
    ]
!

toggleExpand
    "if the item is collapsed, the item is expanded otherwise collapsed"

    self isExpanded ifTrue:[
        self collapse
    ] ifFalse:[
        self expand
    ].
! !

!AbstractHierarchicalItem methodsFor:'accessing-mvc'!

application
    "returns the responsible application or nil"

    |model|

    (model := self model) notNil ifTrue:[
        ^ model application
    ].
    ^ nil
!

applicationsDo:aOneArgBlock
    "evaluate the block for each dependent application"

    |model|

    (model := self model) notNil ifTrue:[
        model applicationsDo:aOneArgBlock
    ]
!

fetchModel
    "returns the hierachicalList model or nil.
     This is a stupid implementation here, in that the top-item's parent is assumed to
     be the model of the tree, and that is returned.
     This saves a slot in every node, but makes some algorithms O(n*log n) or even O(n^2).
     So be aware of the performance penalty"

    |item next|

    item := self. 
    [ (next := item parentOrModel) notNil ] whileTrue:[
        item := next.
    ].

    item isHierarchicalItem ifFalse:[^ item].
    ^ nil
!

model
    "returns the hierachicalList model or nil.
     This uses fetchModel, which has a stupid implementation here, 
     in that the top-item's parent is assumed to be the model of the tree, and that is returned.
     This saves a slot in every node, but makes some algorithms O(n*log n) or even O(n^2).
     So be aware of the performance penalty or redefined fetchModel"

    ^ self fetchModel

    "Modified (comment): / 28-07-2018 / 13:30:24 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'adding & removing'!

add:aChildItem
    "add a child at end"

    ^ self synchronized:[
        self add:aChildItem beforeIndex:(self children size + 1).
    ].

    "Modified: / 28-07-2018 / 13:35:08 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:23:01 / Stefan Vogel"
!

add:aChildItem after:aChild
    "add an item after an existing item"

    ^ self synchronized:[
        |index|

        index := self identityIndexOf:aChild.
        index == 0 ifTrue:[ self subscriptBoundsError:index ].

        self add:aChildItem beforeIndex:(index + 1).
    ].

    "Modified: / 28-07-2018 / 13:35:18 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:23:26 / Stefan Vogel"
!

add:aChildItem afterIndex:anIndex
    "add an item after an index"

    ^ self add:aChildItem beforeIndex:(anIndex + 1).

    "Modified: / 28-07-2018 / 13:37:32 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:23:39 / Stefan Vogel"
!

add:aChildItem before:aChild
    "add an item before an existing item"

    ^ self synchronized:[
        |index|

        index := self identityIndexOf:aChild.
        index == 0 ifTrue:[ self subscriptBoundsError:index ].

        self add:aChildItem beforeIndex:index.
    ].

    "Modified: / 28-07-2018 / 13:35:47 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:22:28 / Stefan Vogel"
!

add:aChildItem beforeIndex:anIndex
    "add an item before an index"

    aChildItem notNil ifTrue:[
        self addAll:(Array with:aChildItem) beforeIndex:anIndex
    ].
    ^ aChildItem.

    "Modified: / 28-07-2018 / 13:35:01 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:24:07 / Stefan Vogel"
!

add:aChild sortBlock:aBlock
    "add a child sorted"

    ^ self synchronized:[
        self basicAdd:aChild sortBlock:aBlock
    ].

    "Modified: / 28-07-2018 / 13:37:12 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:24:31 / Stefan Vogel"
!

addAll:aList
    "add children at the end"

    ^ self synchronized:[
        self addAll:aList beforeIndex:(self children size + 1)
    ]

    "Modified: / 28-07-2018 / 13:37:25 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:25:09 / Stefan Vogel"
!

addAll:aList before:aChild
    "add an item before an existing item"

    ^ self synchronized:[
        |index|

        index := self identityIndexOf:aChild.
        index == 0 ifTrue:[ self subscriptBoundsError:index ].

        self addAll:aList beforeIndex:index
    ].

    "Modified: / 28-07-2018 / 13:37:53 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:25:18 / Stefan Vogel"
!

addAll:aList beforeIndex:anIndex
    "add children before an index"

    aList notEmptyOrNil ifTrue:[
        self synchronized:[
            self basicAddAll:aList beforeIndex:anIndex
        ]
    ].
    ^ aList

    "Modified: / 28-07-2018 / 13:38:02 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:25:38 / Stefan Vogel"
!

addAll:aList sortBlock:aBlock
    "add children sorted"

    aList notEmptyOrNil ifTrue:[ ^ aList ].

    aBlock isNil ifTrue:[
        self addAll:aList.
    ] ifFalse:[
        self synchronized:[
            aList do:[:el| self basicAdd:el sortBlock:aBlock ]
        ]
    ].
    ^ aList

    "Modified: / 28-07-2018 / 13:38:14 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:26:24 / Stefan Vogel"
!

addAllFirst:aCollectionOfItems
    "add children at the beginning"

    ^ self addAll:aCollectionOfItems beforeIndex:1

    "Modified: / 28-07-2018 / 13:38:24 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:26:47 / Stefan Vogel"
!

addAllLast:aCollectionOfItems
    "add children at the end"

    ^ self addAll:aCollectionOfItems

    "Modified: / 28-07-2018 / 13:38:32 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:26:55 / Stefan Vogel"
!

addFirst:aChildItem
    "add a child at the beginning"

    ^ self add:aChildItem beforeIndex:1.

    "Modified: / 28-07-2018 / 13:38:47 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:27:01 / Stefan Vogel"
!

addLast:anItem
    "add a child at the end"

    ^ self add:anItem

    "Modified: / 28-07-2018 / 13:38:52 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:27:09 / Stefan Vogel"
!

remove
    "remove the item"

    parent notNil ifTrue:[                                      "check whether parent exists"
        parent isHierarchicalItem ifTrue:[parent remove:self]   "parent is HierarchicalItem"
                                 ifFalse:[parent root:nil]      "parent is HierarchicalList"
    ].

    "Modified: / 29-07-2018 / 20:31:18 / Stefan Vogel"
!

remove:aChild
    "remove a child"

    self synchronized:[
        self removeIndex:(self identityIndexOf:aChild)
    ].

    "Modified: / 28-07-2018 / 13:39:47 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:40:47 / Stefan Vogel"
!

removeAll
    "remove all children.
     Returns the receiver."

    self synchronized:[
        |size|

        (size := children size) ~~ 0 ifTrue:[
            self removeFromIndex:1 toIndex:size
        ]   
    ]

    "Modified (comment): / 28-07-2018 / 13:44:24 / Claus Gittinger"
!

removeAll:aList
    "remove all children in the collection.
     Returns the receiver."

    aList size ~~ 0 ifTrue:[
        self synchronized:[
            aList do:[:el|
                |index|

                (index := self identityIndexOf:el) ~~ 0 ifTrue:[
                    self removeIndex:index
                ]
            ]
        ]
    ].

    "Modified (format): / 28-07-2018 / 13:44:51 / Claus Gittinger"
    "Modified: / 29-07-2018 / 20:39:01 / Stefan Vogel"
!

removeAllIdentical:aList
    "remove all children in the collection"

    ^ self removeAll:aList.

    "Created: / 20-09-2010 / 09:43:06 / sr"
    "Modified: / 28-07-2018 / 13:41:48 / Claus Gittinger"
!

removeFromIndex:startIndex
    "remove the children from startIndex up to end of children.
     Returns the receiver."

    self synchronized:[
        self removeFromIndex:startIndex toIndex:(children size)
    ].

    "Modified: / 28-07-2018 / 13:44:07 / Claus Gittinger"
!

removeFromIndex:startIndex toIndex:stopIndex
    "remove the children from startIndex up to and including
     the child under stopIndex.
     Returns the receiver."

    self synchronized:[
        |nrOfChildren|

        nrOfChildren := children size.

        (startIndex <= stopIndex and:[startIndex <= nrOfChildren]) ifTrue:[
            self basicRemoveFromIndex:startIndex toIndex:(stopIndex min:nrOfChildren)
        ].
        (children isEmptyOrNil and:[self clearExpandedWhenLastChildWasRemoved]) ifTrue:[
            self setExpanded:false.
        ].
    ].

    "Modified: / 28-07-2018 / 13:45:07 / Claus Gittinger"
    "Modified: / 02-08-2018 / 15:51:19 / Stefan Vogel"
!

removeIndex:anIndex
    "remove the child at an index"

    anIndex > 0 ifTrue:[
        self removeFromIndex:anIndex toIndex:anIndex
    ]
! !

!AbstractHierarchicalItem methodsFor:'basic adding & removing'!

basicAdd:aChild sortBlock:aBlockOrNil
    "add a child sorted.
     Not synchronized - should only be called internally
     within a synchronized region."

    |size list|

    size := children size.
    list := Array with:aChild.

    (aBlockOrNil notNil and:[size ~~ 0]) ifTrue:[
        children keysAndValuesDo:[:i :el|
            (aBlockOrNil value:aChild value:el) ifTrue:[
                self basicAddAll:list beforeIndex:i.
                ^ aChild
            ]
        ]
    ].

    "insert at the end"
    self basicAddAll:list beforeIndex:(size + 1).
    ^ aChild.

    "Modified (comment): / 28-07-2018 / 13:46:12 / Claus Gittinger"
    "Modified (comment): / 01-08-2018 / 16:07:16 / Stefan Vogel"
!

basicAddAll:aList beforeIndex:anIndex
    "add children before an index.
     Not synchronized - should only be called internally
     within a synchronized region."

    |coll model notify index size|

    size := children size.

    anIndex == 1 ifTrue:[
        notify := self
    ] ifFalse:[
        anIndex > size ifTrue:[
            anIndex > (1 + size) ifTrue:[
                ^ self subscriptBoundsError:index
            ].
            notify := self at:size
        ] ifFalse:[
            notify := nil
        ]
    ].

    aList do:[:eachItem| eachItem parent:self ].

    size == 0 ifTrue:[
        children := OrderedCollection withAll:aList.
    ] ifFalse:[
        children isArray ifTrue:[
            children := children asOrderedCollection
        ].
        children addAll:aList beforeIndex:anIndex.
    ].

    (model := self model) isNil ifTrue:[
        ^ aList
    ].

    self isExpanded ifFalse:[
        notify notNil ifTrue:[
            notify changed
        ].
        ^ aList
    ].
    (index := self listIndex) isNil ifTrue:[
        ^ aList
    ].

    children from:1 to:(anIndex - 1) do:[:anItem|
        index := 1 + index + anItem numberOfVisibleChildren
    ].

    coll := OrderedCollection new.
    aList do:[:anItem|
        coll add:anItem.
        anItem addVisibleChildrenTo:coll.
    ].
    model itemAddAll:coll beforeIndex:(index + 1).

    notify notNil ifTrue:[
        notify changed
    ].
    ^ aList

    "Modified (comment): / 28-07-2018 / 13:46:06 / Claus Gittinger"
    "Modified: / 01-08-2018 / 15:31:54 / Stefan Vogel"
    "Modified (format): / 02-08-2018 / 15:52:16 / Stefan Vogel"
!

basicRemoveFromIndex:startIndex toIndex:stopIndex
    "remove the children from startIndex up to and including
     the child under stopIndex.
     Not synchronized - should only be called internally
     within a synchronized region."

    |model notify
     index "{ Class:SmallInteger }"
     start "{ Class:SmallInteger }"
     stop  "{ Class:SmallInteger }"
     size  "{ Class:SmallInteger }"
    |
    size  := self children size.
    stop  := stopIndex.
    start := startIndex.

    (stop <= size and:[start between:1 and:stop]) ifFalse:[
        ^ self subscriptBoundsError:index
    ].
    start == 1 ifTrue:[
        notify := self
    ] ifFalse:[
        stop == size ifTrue:[
            notify := self at:(start - 1)
        ] ifFalse:[
            notify := nil
        ]
    ].

    (model := self model) notNil ifTrue:[
        index := model identityIndexOf:(children at:start).
        size  := stop - start + 1.
    ] ifFalse:[
        index := 0
    ].

    children from:start to:stop do:[:aChild|
        index ~~ 0 ifTrue:[
            size := size + aChild numberOfVisibleChildren
        ].
        aChild parent:nil
    ].
    children removeFromIndex:start toIndex:stop.

    index ~~ 0 ifTrue:[
        model itemRemoveFromIndex:index toIndex:(index + size - 1)
    ].
    notify notNil ifTrue:[
        notify changed
    ].

    "Modified (format): / 28-07-2018 / 13:46:41 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'change & update'!

changed:what with:anArgument
    "the item changed; raise change notification
        #icon           icon is modified; height and width are unchanged
        #hierarchy      collapsed/expanded; height and width are unchanged
        #redraw         redraw but height and width are unchanged
        .......         all others: the height and width are reset
    "
    
    |model|

    (what ~~ #redraw 
     and:[what ~~ #hierarchy 
     and:[what ~~ #icon 
     and:[self class doResetExtentOnChange]]]) ifTrue:[
        self makeWidthAndHeightUnknown
    ].

    (model := self model) notNil ifTrue:[
        self synchronized:[
            model itemChanged:what with:anArgument from:self.
        ].
    ].
    super changed:what with:anArgument

    "Modified: / 24-11-2010 / 17:21:20 / cg"
    "Modified: / 30-07-2018 / 15:58:18 / Stefan Vogel"
    "Modified (format): / 15-08-2018 / 12:37:03 / Claus Gittinger"
!

childrenOrderChanged
    "called if the order of the children changed by a user
     operation. Update the model and raise a change notification for
     each item which has changed its position
     triggered by the user operation !!"

    |model visStart list|

    self isExpanded   ifFalse:[ ^ self ].       "/ not expanded
    children size > 1 ifFalse:[ ^ self ].

    model := self model.
    model isNil ifTrue:[^ self].                       "/ no model

    visStart := model identityIndexOf:self.
    visStart == 0 ifTrue:[
        model root ~~ self ifTrue:[ ^ self ].
        "/ I'am the root but switched of by setting #showRoot to false
    ].

    self synchronized:[
        list := OrderedCollection new.
        self addVisibleChildrenTo:list.

        list do:[:el|
            visStart := visStart + 1.

            (model at:visStart ifAbsent:el) ~~ el ifTrue:[
                model at:visStart put:el
            ].
        ]
    ].

    "Modified (format): / 28-07-2018 / 13:47:27 / Claus Gittinger"
!

fontChanged
    "called if the font has changed.
     Clear the precomputed width and height"

    self makeWidthAndHeightUnknown.

    children notNil ifTrue:[
        children do:[:el| el fontChanged].
    ].

    "Modified: / 02-08-2018 / 15:53:52 / Stefan Vogel"
!

hierarchyChanged
    "hierarchy changed; optimize redrawing"

    self changed:#hierarchy with:nil
!

iconChanged
    "icon changed; optimize redrawing"

    self changed:#icon with:nil
!

invalidate
    "force redrawing of me"

    self changed:#redraw with:nil

    "Created: / 15-08-2018 / 12:38:48 / Claus Gittinger"
!

labelChanged
    "called if the label has changed.
     Clear the precomputed width and height, and invalidate"

    self makeWidthAndHeightUnknown.
    self invalidate

    "Created: / 17-01-2011 / 17:43:42 / cg"
    "Modified: / 15-08-2018 / 12:43:20 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'enumerating'!

allExpandedItemsDo:aBlock
    "recursively enumerate all expanded nodes 
     (depth first; parent before children)"
    
    self recursiveDo:[:each |
        each isExpanded ifTrue:[
            aBlock value:each.
        ].
    ].

    "Created: / 16-10-2017 / 14:19:00 / cg"
!

collect:aBlock
    "for each child in the receiver (non recursive), evaluate the argument, aBlock
     and return a new collection with the results"

    |coll|

    coll := OrderedCollection new.
    self do:[:el| coll add:(aBlock value:el) ].
    ^ coll

    "Modified (comment): / 25-11-2016 / 08:42:19 / cg"
!

contains:aBlock
    "evaluate aOneArgBlock for each of the receiver's children (non recursive).
     Return true and skip remaining elements, if aBlock ever returns true, 
     otherwise return false"

    self do:[:el | (aBlock value:el) ifTrue:[^ true] ].
    ^ false

    "Modified (comment): / 25-11-2016 / 08:43:31 / cg"
!

do:aOneArgBlock
    "evaluate a block for each child (non recursive)"

    self from:1 to:nil do:aOneArgBlock

    "Modified (comment): / 25-11-2016 / 08:43:36 / cg"
!

from:startIndex do:aOneArgBlock
    "evaluate a block on each child starting with the
     child at startIndex to the end."

    self from:startIndex to:nil do:aOneArgBlock

    "Modified (comment): / 25-11-2016 / 07:45:55 / cg"
!

from:startIndex reverseDo:aOneArgBlock
    "evaluate a block on each child (non recursive) starting at end to the startIndex"

    self from:startIndex to:nil reverseDo:aOneArgBlock

    "Modified (comment): / 25-11-2016 / 08:43:42 / cg"
!

from:startIndex to:endIndex do:aOneArgBlock
    "evaluate a block on each child (non recursive),
     starting with the child at startIndex to the endIndex (last index if nil).
     Answer the value of the block executed on the last element."

    ^ self synchronized:[
        self nonCriticalFrom:startIndex to:endIndex do:aOneArgBlock
    ].

    "Modified (comment): / 25-11-2016 / 08:43:51 / cg"
    "Modified: / 28-07-2018 / 13:48:49 / Claus Gittinger"
    "Modified: / 02-08-2018 / 16:04:04 / Stefan Vogel"
!

from:startIndex to:endIndex reverseDo:aOneArgBlock
    "evaluate a block on each child (non recursive),
     starting with the child at endIndex to the startIndex."

    self synchronized:[
        self size >= startIndex ifTrue:[
            self nonCriticalFrom:startIndex to:endIndex reverseDo:aOneArgBlock
        ].
    ].

    "Modified (comment): / 25-11-2016 / 08:43:58 / cg"
    "Modified: / 28-07-2018 / 13:49:17 / Claus Gittinger"
!

keysAndValuesDo:aTwoArgBlock
    "evaluate the argument, aBlock for every child (non recursive),
     passing both index and element as arguments."

    |key|

    key := 1.

    self do:[:el|
        el value:key value:el.
        key := key + 1.
    ].

    "Modified (comment): / 25-11-2016 / 08:44:03 / cg"
    "Modified: / 28-07-2018 / 13:49:35 / Claus Gittinger"
!

keysAndValuesReverseDo:aTwoArgBlock
    "evaluate the argument, aBlock in reverse order for every child (non recursive), 
     passing both index and element as arguments."

    self synchronized:[
        self size == 0 ifFalse:[
            self nonCriticalKeysAndValuesReverseDo:aTwoArgBlock
        ].
    ].

    "Modified (comment): / 25-11-2016 / 08:44:06 / cg"
    "Modified: / 28-07-2018 / 13:50:14 / Claus Gittinger"
!

recursiveCollect:aBlock
    "for each child in the receiver, evaluate the argument, aBlock
     and return a new collection with the results.
     Warning: this only enumerates already visible child elements
     i.e. any collapsed items are not visited."

    |coll|

    coll := OrderedCollection new.
    self recursiveDo:[:el| coll add:(aBlock value:el) ].
    ^ coll

    "Modified (comment): / 25-11-2016 / 08:43:20 / cg"
!

recursiveDo:aOneArgBlock
    "evaluate a block on each item and all the sub-items.
     Warning: this only enumerates already visible child elements
     i.e. any collapsed items are not visited."

    self do:[:aChild| "/ notice: do does it synchronized
        aOneArgBlock value:aChild.
        aChild nonCriticalRecursiveDo:aOneArgBlock
    ].

    "Modified (comment): / 25-11-2016 / 08:43:04 / cg"
    "Modified (format): / 28-07-2018 / 13:56:02 / Claus Gittinger"
!

recursiveReverseDo:aOneArgBlock
    "evaluate a block on each item and all the sub-items;
     proccesing children in reverse direction.
     Warning: this only enumerates already visible child elements
     i.e. any collapsed items are not visited."

    self reverseDo:[:aChild| "/ notice: reverseDo does it synchronized
        aChild nonCriticalRecursiveReverseDo:aOneArgBlock.
        aOneArgBlock value:aChild.
    ].

    "Modified (comment): / 25-11-2016 / 08:43:09 / cg"
    "Modified (format): / 28-07-2018 / 13:56:34 / Claus Gittinger"
!

recursiveSelect:aBlock
    "return a new collection with all children and subChildren from the receiver, 
     for which the argument aBlock evaluates to true.
     Warning: this only enumerates already visible child elements
     i.e. any collapsed items are not visited."

    |coll|

    coll := OrderedCollection new.
    self recursiveDo:[:el| (aBlock value:el) ifTrue:[coll add:el] ].
    ^ coll

    "Modified (comment): / 25-11-2016 / 08:43:12 / cg"
!

reverseDo:aOneArgBlock
    "evaluate a block on each child (non recursive) in reverse direction"

    self from:1 reverseDo:aOneArgBlock

    "Modified (comment): / 25-11-2016 / 08:44:12 / cg"
!

select:aBlock
    "return a new collection with all items from the receiver (non recursive), 
     for which the argument aBlock evaluates to true."

    |coll|

    coll := OrderedCollection new.
    self do:[:el| (aBlock value:el) ifTrue:[coll add:el] ].
    ^ coll

    "Modified (comment): / 25-11-2016 / 08:44:20 / cg"
!

withAllDo:aOneArgBlock
    "recursively evaluate aOneArgBlock on each item and subitem including self"

    aOneArgBlock value:self.

    self do:[:el| "/ do does it synchronized
        aOneArgBlock value:el.
        el nonCriticalRecursiveDo:aOneArgBlock.
    ].

    "Modified (comment): / 25-11-2016 / 08:44:49 / cg"
    "Modified (format): / 28-07-2018 / 15:44:45 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'enumerating parents'!

allParents
    "return a collection of all parents (in parent, grandparent, ... order)"

    |parents|

    parents := OrderedCollection new.
    self parentsDo:[:p | parents add:p].
    ^ parents.

    "Created: / 16-10-2017 / 13:08:45 / cg"
!

parentsDetect:aBlock
    "find the first parent, for which evaluation of the block returns
     true; if none does so, report an error"

    ^ self parentsDetect:aBlock ifNone:[self errorNotFound]
!

parentsDetect:aBlock ifNone:anExceptionBlock
    "find the first parent, for which evaluation of the block returns
     true; if none does so, return the evaluation of anExceptionBlock"

    |prnt|

    prnt := self.

    self synchronized:[
        [(prnt := prnt parent) notNil and:[prnt isHierarchicalItem]] whileTrue:[
            (aBlock value:prnt) ifTrue:[^ prnt]
        ]
    ].
    ^ anExceptionBlock value

    "Modified: / 28-07-2018 / 13:51:06 / Claus Gittinger"
!

parentsDo:aBlock
    "evaluate a block for each parent"

    |prnt|

    prnt := self.

    self synchronized:[
        [(prnt := prnt parent) notNil and:[prnt isHierarchicalItem]] whileTrue:[
            aBlock value:prnt
        ]
    ].

    "Modified: / 28-07-2018 / 13:51:22 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'initialization'!

initialize
    self setExpanded:false
! !

!AbstractHierarchicalItem methodsFor:'private'!

addVisibleChildrenTo:aList
    "add all visible children and sub-children to the list"

    self isExpanded ifFalse:[^ self].

    self from:1 to:nil do:[:el|
        aList add:el.
        el addVisibleChildrenTo:aList.
    ].

    "Modified: / 28-07-2018 / 13:58:20 / Claus Gittinger"
!

clearExpandedWhenLastChildWasRemoved
    "https://expeccoalm.exept.de/D227397 
     Do not set #isExpanded to false just because #children is empty (may children appear 'again' later).
     Do modify #isExpanded ONLY when a user presses the expand/collapse toggle, otherwise #isExpanded should be persistent.

     The user's preference if the item is expanded or collapsed should be kept,    
     regardless if there are chilren or not (even regardless anything else).
     All other related things, like the drawing in case for #isExpanded is true and children is empty, 
     has to be solved within the drawing (or within any feature requesting this information)"

    ^ false
!

criticalDo:aBlock
    self synchronized:aBlock

    "Modified: / 23-07-2018 / 13:25:11 / Stefan Vogel"
    "Modified: / 28-07-2018 / 15:51:23 / Claus Gittinger"
!

listIndex
    "returns the visible index or nil; for a non-visible root, 0 is returned"

    |index model|

    (model := self model) notNil ifTrue:[
        index := model identityIndexOf:self.

        (index ~~ 0 or:[parent == model]) ifTrue:[
            ^ index
        ]
    ].
    ^ nil
!

numberOfVisibleChildren
    "returns the number of all visible children including subchildren"

    |count|

    self isExpanded ifFalse:[^ 0].

    count := 0.
    self from:1 to:nil do:[:el|
        count := 1 + count + (el numberOfVisibleChildren)
    ].
    ^ count

    "Modified: / 28-07-2018 / 13:55:36 / Claus Gittinger"
!

parentOrModel
    "returns the parent without checking for item or model"

    ^ parent
!

synchronized:aBlock
    |model|

    (model := self model) notNil ifTrue:[
        ^ model synchronized:aBlock
    ] ifFalse:[
        self breakPoint:#cg. "/ should not happen
        ^ aBlock value
    ]

    "Created: / 28-07-2018 / 13:58:54 / Claus Gittinger"
    "Modified: / 29-07-2018 / 21:02:22 / Stefan Vogel"
! !

!AbstractHierarchicalItem methodsFor:'private-displaying'!

displayLabel:aLabel h:lH on:aGC x:x y:y h:h
    <resource: #obsolete>
 
    "display the label at x@y"

    "/ obsolete - left in for backward compatibility
    self displayLabel:aLabel h:lH on:aGC x:x y:y h:h isHighlightedAsSelected:false
!

displayLabel:aLabel h:lH on:aGC x:x y:y h:h isHighlightedAsSelected:isHighlightedAsSelected
    "display the label at x@y"

    |y0|

    lH ~~ 0 ifTrue:[
        y0 := y - ((lH + 1 - h) // 2).
        y0 := y0 + (aLabel ascentOn:aGC).

        (aLabel isString not
         or:[(aLabel includes:(Character cr)) not]) ifTrue:[
            |labelShown|

            labelShown := aLabel.
            isHighlightedAsSelected ifTrue:[
                (aLabel isText and:[aLabel hasChangeOfEmphasis]) ifTrue:[
                    labelShown := Text string:aLabel string emphasisCollection:aLabel emphasis asArray.
                    labelShown emphasisAllRemove:#color.
                ].
            ].
            labelShown displayOn:aGC x:x y:y0.
            ^ self.
        ].

        aLabel asCollectionOfLines do:[:el|
            |labelShown|

            labelShown := el.
            isHighlightedAsSelected ifTrue:[
                (el isText and:[el hasChangeOfEmphasis]) ifTrue:[
                    labelShown := Text string:el string emphasisCollection:el emphasis asArray.
                    labelShown emphasisAllRemove:#color.
                ].
            ].
            labelShown displayOn:aGC x:x y:y0.
            y0 := y0 + (el heightOn:aGC)
        ]
    ].
!

heightOf:aLabel on:aGC
    "returns the height of the label or 0"

    |h|

    aLabel isSequenceable ifFalse:[
        ^ aLabel notNil ifTrue:[aLabel heightOn:aGC] ifFalse:[0]
    ].

    aLabel isString ifFalse:[
        h := 0.
        aLabel do:[:el|h := h max:(self heightOf:el on:aGC)].
        ^ h
    ].

    h := 1 + (aLabel occurrencesOf:(Character cr)).
    ^ h * (aGC font height)
!

widthOf:aLabel on:aGC
    "returns the height of the label or 0"

    |w|

    aLabel isSequenceable ifFalse:[
        ^ aLabel notNil ifTrue:[aLabel widthOn:aGC] ifFalse:[0]
    ].

    aLabel isString ifFalse:[
        w := -5.
        aLabel do:[:el|w := w + 5 + (self widthOf:el on:aGC)].
        ^ w
    ].

    (aLabel indexOf:(Character cr)) == 0 ifTrue:[
        ^ aLabel widthOn:aGC
    ].

    w := 0.
    aLabel asCollectionOfLines do:[:el|w := w max:(el widthOn:aGC)].
    ^ w
! !

!AbstractHierarchicalItem methodsFor:'private-enumerating'!

nonCriticalDo:aOneArgBlock
    "evaluate a block noncritical for each child.
     Not synchronized - should only be called internally
     within a synchronized region."

    ^ self nonCriticalFrom:1 to:nil do:aOneArgBlock

    "Modified (comment): / 28-07-2018 / 13:52:18 / Claus Gittinger"
!

nonCriticalFrom:startIndex to:endIndex do:aOneArgBlock
    "evaluate a block noncritical for each child starting with the
     child at startIndex to the endIndex (if nil to end of list).
     Not synchronized - should only be called internally
     within a synchronized region.
     Answer the value of the block executed on the last element."

    |list size resp|

    list := self children.
    size := list size.

    startIndex > size ifTrue:[^ nil].
    resp := nil.

    endIndex notNil ifTrue:[
        size := size min:endIndex
    ].
    startIndex to:size do:[:i| 
        |item|

        item := list at:i ifAbsent:nil.
        item isNil ifTrue:[^ resp].
        resp := aOneArgBlock value:item.
    ].
    ^ resp

    "Modified (comment): / 28-07-2018 / 13:52:24 / Claus Gittinger"
    "Modified (comment): / 02-08-2018 / 16:02:13 / Stefan Vogel"
!

nonCriticalFrom:startIndex to:endIndex reverseDo:aOneArgBlock
    "evaluate a block non critical for each child starting with the
     child at endIndex (if nil to end of list) to startIndex.
     Not synchronized - should only be called internally
     within a synchronized region."

    |list size resp|

    list := self children.
    size := list size.
    resp := nil.

    endIndex notNil ifTrue:[
        size := size min:endIndex
    ].
    size to:startIndex by:-1 do:[:i| 
        |item|

        item := list at:i ifAbsent:nil.
        item isNil ifTrue:[^ resp].
        resp := aOneArgBlock value:item.
    ].
    ^ resp

    "Modified (comment): / 28-07-2018 / 13:52:30 / Claus Gittinger"
!

nonCriticalKeysAndValuesReverseDo:aOneArgBlock
    "evaluate the argument, aBlock in reverse order for every
     child, passing both index and element as arguments.
     Not synchronized - should only be called internally
     within a synchronized region."

    |list size resp|

    list := self children.
    size := list size.
    resp := nil.

    size to:1 by:-1 do:[:i| 
        |item|

        item := list at:i ifAbsent:nil.
        item isNil ifTrue:[^ resp].
        resp := aOneArgBlock value:i value:item.
    ].
    ^ resp

    "Modified (comment): / 28-07-2018 / 13:52:35 / Claus Gittinger"
!

nonCriticalRecursiveDo:aOneArgBlock
    "evaluate the block non critical for each item and all the sub-items.
     Not synchronized - should only be called internally
     within a synchronized region."

    self nonCriticalFrom:1 to:nil do:[:eachChild|
        aOneArgBlock value:eachChild.
        eachChild nonCriticalRecursiveDo:aOneArgBlock
    ].

    "Modified (comment): / 28-07-2018 / 13:52:41 / Claus Gittinger"
!

nonCriticalRecursiveReverseDo:aOneArgBlock
    "evaluate the block non critical for each item and all the sub-items;
     proccesing children in reverse direction.
     Not synchronized - should only be called internally
     within a synchronized region."

    self nonCriticalFrom:1 to:nil reverseDo:[:eachChild|
        eachChild nonCriticalRecursiveReverseDo:aOneArgBlock.
        aOneArgBlock value:eachChild.
    ].

    "Modified (comment): / 28-07-2018 / 13:52:46 / Claus Gittinger"
!

nonCriticalRecursiveSort:aSortBlock
    "evaluate a block noncritical for each child.
     Not synchronized - should only be called internally
     within a synchronized region."

    |unsorted sorted|

    unsorted := children.

    unsorted size ~~ 0 ifTrue:[
        sorted := unsorted sort:aSortBlock.
        sorted do:[:el| el nonCriticalRecursiveSort:aSortBlock ].
        children := sorted.
    ].

    "Modified (comment): / 28-07-2018 / 13:52:50 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'private-hierarchy'!

recursiveSetCollapsed
    "collapse all children and sub-children without notifications"

    self synchronized:[
        self recursiveSetCollapsedHelper
    ]

    "Modified: / 28-07-2018 / 13:56:58 / Claus Gittinger"
!

recursiveSetCollapsedHelper
    "private helper.
     collapse all children and sub-children without notifications.
     Helper; not synchronized - should only be called internally
     within a synchronized region."

    self setExpanded:false.

    "/ do not call #size: children will be autoloaded !!!!
    self nonCriticalFrom:1 to:nil do:[:eachChild| 
        eachChild canRecursiveCollapse ifTrue:[
            eachChild recursiveSetCollapsedHelper
        ]
    ].

    "Modified (comment): / 28-07-2018 / 13:57:40 / Claus Gittinger"
    "Modified: / 02-08-2018 / 16:07:35 / Stefan Vogel"
!

recursiveSetExpandedAndAddToList:aList
    "expand all children and sub-children without notifications;
     add children to list"

    self synchronized:[
        self recursiveSetExpandedAndAddToListHelper:aList
    ].

    "Modified: / 28-07-2018 / 13:59:52 / Claus Gittinger"
!

recursiveSetExpandedAndAddToListHelper:aList
    "private helper.
     expand all children and sub-children without notifications; adds children to aList
     Helper; not synchronized - should only be called internally
     within a synchronized region."

    self setExpanded:true.

    self nonCriticalFrom:1 to:nil do:[:eachChild|
        aList add:eachChild.

        eachChild canRecursiveExpand ifTrue:[
            eachChild recursiveSetExpandedAndAddToListHelper:aList.
        ].
    ].

    "Modified (comment): / 28-07-2018 / 14:00:04 / Claus Gittinger"
! !

!AbstractHierarchicalItem methodsFor:'private-to be redefined'!

fetchChildren
    ^ nil
!

heightOn:aGC
    "return the height of the receiver, if it is to be displayed on aGC"

    self subclassResponsibility.

    "/ could in theory compute it for every draw operation;
    "/ in practice, this is not a good idea, as it could make drawing very slow
    "/ (unless that info can be easily computed)

    "/ so we force programmers to think about that issue and redefine this method
    "/ as required. If compuation is really cheap, it can be redefined as:
    "/ ^ self heightOf:(self label) on:aGC
!

isExpanded
    "returns true if the item is expanded"

    self subclassResponsibility
!

makeWidthAndHeightUnknown
    "invalidate any cached with/height information"

    "see comments in widthOn/heightOn"
    self subclassResponsibility
!

setExpanded:aBoolean
    "set expanded flag without any computation or notification.
     It is left to the subclasses responsibility, where this expanded state is stored;
     could be in the model (as a list of expanded items), in the item itself (as boolean flag),
     or somewhere else.
     For huge trees, it may make sense to not store the expanded flag in a slot 
     (in order to save space). See CompactHierarchicalItem as a clever example of how it can be
     stored without ANY additional space requirements."

    self subclassResponsibility
!

widthOn:aGC
    "return the width of the receiver, if it is to be displayed on aGC"

    self subclassResponsibility.

    "/ could in theory compute it for every draw operation;
    "/ in practice, this is not a good idea, as it could make drawing very slow
    "/ (unless that info can be easily computed)

    "/ so we force programmers to think about that issue and redefine this method
    "/ as required. If compuation is really cheap, it can be redefined as:
    "/ ^ self widthOf:(self label) on:aGC
! !

!AbstractHierarchicalItem methodsFor:'protocol-accessing'!

children
    "returns a list of children. When first asked, the list is fetched, if it was
     built lazily.
     *** to optimize: either redefine this or fetchChildren by subClass"

    children isNil ifTrue:[
        children := self fetchChildren
    ].
    ^ children
!

icon
    "returns the icon or nil;
     *** to optimize:redefine by subClass"

    |model|

    (model := self model) notNil ifTrue:[
        ^ model iconFor:self
    ].
    ^ nil
!

label
    "returns the label displayed on aGC;
     *** to optimize:redefine by subClass"

    |model|

    (model := self model) notNil ifTrue:[
        ^ model labelFor:self
    ].
    ^ nil
!

middleButtonMenu
    "returns the items middleButtonMenu or nil if no menu is defined.
     If nil is returned, the view is asked for a menu."

    <resource: #programMenu>

    |model|

    (model := self model) notNil ifTrue:[
        ^ model middleButtonMenuFor:self
    ].
    ^ nil
!

recursiveSortChildren:aSortBlock
    self synchronized:[
        |children|

        (children := self children) notEmptyOrNil ifTrue:[
            children sort:aSortBlock.
            children do:[:aChild| aChild recursiveSortChildren:aSortBlock ]
        ]
    ].

    "Modified: / 28-07-2018 / 14:01:15 / Claus Gittinger"
!

sortChildren:aSortBlock
    "sort the children inplace using the 2-arg block sortBlock for comparison"

    <resource: #obsolete>
    self obsoleteMethodWarning:'use #sort:'.
    self sort:aSortBlock.
! !

!AbstractHierarchicalItem methodsFor:'protocol-displaying'!

displayIcon:anIcon atX:x y:y on:aGC
    "called to draw the icon - can be redefined to manipulate the icon"

    anIcon displayOn:aGC x:x y:y.
!

displayOn:aGC x:x y:y h:h
    <resource: #obsolete>
    "draw the receiver in the graphicsContext, aGC"

    "/ obsolete - left here for backward compatibility
    self displayOn:aGC x:x y:y h:h isHighlightedAsSelected:false
!

displayOn:aGC x:x y:y h:h isHighlightedAsSelected:isHighlightedAsSelected
    "draw the receiver in the graphicsContext, aGC"

    |label
     x0 "{ Class:SmallInteger }"|

    label := self label.
    label isNil ifTrue:[^ self].

    label isNonByteCollection ifFalse:[
        ^ self displayLabel:label h:(self heightOn:aGC) on:aGC x:x y:y h:h isHighlightedAsSelected:isHighlightedAsSelected
    ].

    x0 := x.
    label do:[:el|
        el notNil ifTrue:[
            self displayLabel:el h:(self heightOf:el on:aGC) on:aGC x:x0 y:y h:h isHighlightedAsSelected:isHighlightedAsSelected.
            x0 := x0 + 5 + (el widthOn:aGC).
        ].
    ]
! !

!AbstractHierarchicalItem methodsFor:'protocol-event processing'!

processButtonPress:button visibleX:visX visibleY: visY on: view
    "A mouse button was pressed on myself. The visX/visY coordinates
     are relative to the viewOrigin.

     If this method returns TRUE, the other method
     #processButtonPress:x:y:on: IS NOT CALLED !!!!!!
    "

    ^false

    "Created: / 18-04-2013 / 09:56:29 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (comment): / 18-04-2013 / 11:04:38 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

processButtonPress:button x:x y:y
    "a mouse button was pressed in my label.
     Return true, if the event is eaten (ignored by the gc).
     By default, false is returned (should be handled by the gc)."

    ^ false
!

processButtonPress:button x:x y:y on:aGC
    "a mouse button was pressed in my label.
     Return true, if the event is eaten (ignored by the gc).
     By default, false is returned (should be handled by the gc)."

    ^ self processButtonPress:button x:x y:y
!

processButtonPressOnIcon:button on:aGC
    "a mouse button was pressed in my icon.
     Return true, if the event is eaten (ignored by the gc).
     By default, false is returned (should be handled by the gc)."

    ^ false
! !

!AbstractHierarchicalItem methodsFor:'protocol-monitoring'!

monitoringCycle
    "called every 'n' seconds by the model, if the monitoring
     cycle is enabled. The item can perform some checks, ..
     **** can be redefined by subclass to perform some actions
    "
! !

!AbstractHierarchicalItem methodsFor:'protocol-queries'!

canCollapse
    "called before collapsing the item; can be redefined
     by subclass to omit the collapse operation"

    ^ self isExpanded
!

canExpand
    "called before expanding the item; can be redefined
     by subclass to omit the expand operation"

    ^ self hasChildren
!

canRecursiveCollapse
    "called before collapsing the item; can be redefined
     by subclass to omit the collapse operation "

    ^ self canCollapse
!

canRecursiveExpand
    "called before expanding the item; can be redefined
     by subclass to omit the collapse operation"

    ^ self canExpand
!

drawHorizontalLineUpToText
    "draw the horizontal line for the selected item up to the text
     or (by default) to the start of the vertical line; 
     only used by the hierarchical view
    "
    ^ false

    "Modified (comment): / 12-06-2018 / 10:41:54 / Claus Gittinger"
!

hasChildren
    "checks whether the item has children; 
     the list needs not to be loaded yet( example. FileDirectory ).
     *** to optimize: redefine in subClass"

    ^ self children notEmptyOrNil

    "Modified: / 08-03-2018 / 19:37:45 / stefan"
!

hasIndicator
    "on default the indicator is drawn if the item has children"

    ^ self hasChildren
!

isSelectable
    "returns true if the item is selectable. Can be redefined in subclasses"

    ^ true
!

string
    "access the printable string used for stepping through a list
     searching for an entry starting with a character.
     *** to optimize:redefine by subClass"

    |label|

    (label := self label) notNil ifTrue:[
        label isString      ifTrue:[ ^ label string ].
        label isImageOrForm ifTrue:[ ^ nil ].

        label isSequenceable ifFalse:[
            ^ label perform:#string ifNotUnderstood:nil
        ].

        label do:[:el||s|
            (el notNil and:[el isImageOrForm not]) ifTrue:[
                s := el perform:#string ifNotUnderstood:nil.
                s notNil ifTrue:[^ s].
            ]
        ]
    ].
    ^ nil
! !

!AbstractHierarchicalItem methodsFor:'queries'!

isChildOf:anItem
    "returns true if the item is a child (or grandchild or ...) of anItem"

    self synchronized:[
        |item|

        item := self.

        [anItem ~~ item] whileTrue:[
            ((item := item parent) notNil and:[item isHierarchicalItem]) ifFalse:[
                ^ false
            ]
        ].
    ].
    ^ true

    "Modified: / 28-07-2018 / 14:02:32 / Claus Gittinger"
    "Modified (comment): / 12-08-2018 / 12:49:16 / Claus Gittinger"
!

isCollapsed
    "returns true if the item is collapsed"

    ^ self isExpanded not
!

isDirectoryItem
    ^ false

    "Created: / 23-06-2006 / 12:47:05 / fm"
    "Modified: / 23-02-2007 / 12:04:23 / User"
!

isHierarchicalItem
    "used to decide if the parent is a hierarchical item or the model"

    ^ true
!

isRealChildOf:anItem
    "returns true if the item is a child of anItem"

    self synchronized:[
        |item|
        item := self parent.

        [item notNil] whileTrue:[
            item == anItem ifTrue:[^ true].
            item := item parent.
        ].
    ].
    ^ false

    "Modified: / 28-07-2018 / 14:03:22 / Claus Gittinger"
!

isRootItem
    "returns true if the item is the root item"

    ^ parent isHierarchicalItem not
!

size
    "return the number of children"

    ^ self children size
! !

!AbstractHierarchicalItem methodsFor:'searching'!

detect:aOneArgBlock
    "find the first child (not recursive), for which evaluation of the block returns
     true; if none does so, report an error"

    ^ self detect:aOneArgBlock ifNone:[self errorNotFound]

    "Modified (comment): / 25-11-2016 / 08:40:16 / cg"
!

detect:aOneArgBlock ifNone:exceptionValue
    "find the first child (not recursive), for which evaluation of the block returns
     true; if none does so, return the value of anExceptionValue"

    self do:[:el| 
        (aOneArgBlock value:el) ifTrue:[^ el] 
    ].
    ^ exceptionValue value

    "Modified (comment): / 25-11-2016 / 08:40:18 / cg"
!

detectLast:aOneArgBlock
    "find the last child (not recursive), for which evaluation of the block returns
     true; if none does so, an exception is raised"

    ^ self detectLast:aOneArgBlock ifNone:[self errorNotFound]

    "Modified (comment): / 25-11-2016 / 08:40:21 / cg"
!

detectLast:aOneArgBlock ifNone:anExceptionValue
    "find the last child (not recursive), for which evaluation of the block returns
     true; if none does so, return the value of anExceptionValue"

    self reverseDo:[:el| (aOneArgBlock value:el) ifTrue:[^ el] ].
    ^ anExceptionValue value

    "Modified (comment): / 25-11-2016 / 08:40:23 / cg"
!

findChildByLabelPath:aLabelPath
    "recursivly find the child by its label path.
     A label path is an ordered collection (or array) of strings, 
     which are the labels of items (i.e. as retrived by labelPath)."

    |restPath searchedLabel childSearched|

    childSearched := self.
    restPath := aLabelPath.

    [
        |subChild|
        
        restPath isEmptyOrNil ifTrue:[^ childSearched].

        searchedLabel := restPath first string.
    
        subChild := childSearched detect:[:eachSubChild | eachSubChild label string = searchedLabel] ifNone:nil. 
        subChild isNil ifTrue:[^ nil].

        childSearched := subChild.
        restPath := restPath from:2.
    ] loop.

    "Created: / 16-10-2017 / 14:30:11 / cg"
!

findFirst:aOneArgBlock
    "find the first child (not recursive), for which evaluation of the argument, aOneArgBlock
     returns true; return its index or 0 if none detected."


    self keysAndValuesDo:[:i :el| (aOneArgBlock value:el) ifTrue:[^ i] ].
    ^ 0

    "Modified (comment): / 25-11-2016 / 08:40:25 / cg"
!

findLast:aOneArgBlock
    "find the last child (not recursive), for which evaluation of the argument, aOneArgBlock
     returns true; return its index or 0 if none detected."

    self keysAndValuesReverseDo:[:i :el| (aOneArgBlock value:el) ifTrue:[^ i] ].
    ^ 0

    "Modified (format): / 25-11-2016 / 08:40:33 / cg"
!

findNextInTreeForWhich:aBlock
    "does a breadth-first search in the tree starting at the receiver
     i.e. first goes down recursively, then up to the next child of the parent 
     (doing a recursive search there),
     up until all of the top item's children have been visited.
     Returns the found item or nil."

    |item nextItem itemsParent|

    item := self.
    [item notNil] whileTrue:[
        item recursiveDo:[:eachChild |
            (aBlock value:eachChild) ifTrue:[
                ^ eachChild.
            ].
        ].

        itemsParent := item parent.
        nextItem := nil.

        [
            |idx siblings|

            itemsParent isNil ifTrue:[
                ^ nil
            ].
            siblings := itemsParent children.
            idx := siblings identityIndexOf:item.
            nextItem := siblings at:idx+1 ifAbsent:nil.
            nextItem isNil ifTrue:[
                item := itemsParent.
                itemsParent := itemsParent parent.
            ].
        ] doWhile:[ nextItem isNil ].

        item := nextItem.
        (aBlock value:item) ifTrue:[
            ^ item.
        ].
    ].
    ^ nil

    "Created: / 31-07-2018 / 09:01:11 / Claus Gittinger"
!

findPreviousInTreeForWhich:aBlock
    "does a breadth-first search in the tree starting at the receiver
     i.e. first goes down recursively, then up to the next child of the parent 
     (doing a recursive search there),
     up until all of the top item's children have been visited.
     Returns the found item or nil."

    |item itemsParent idx siblings prevItem|

    item := self.
    [item notNil] whileTrue:[
        "/ previous sibling
        itemsParent := item parent.
        itemsParent isNil ifTrue:[
            ^ nil
        ].
        siblings := itemsParent children.
        idx := siblings identityIndexOf:item.
        idx <= 1 ifTrue:[
            "/ continue in parent
            item := itemsParent.
        ] ifFalse:[
            prevItem := siblings at:idx-1 ifAbsent:nil.
            "/ continue in prev sibling's last child
            item := prevItem.
            [item children notEmptyOrNil] whileTrue:[
                item := item children last.
            ]    
        ]. 
        item isNil ifTrue:[^ nil].
        item isRootItem ifTrue:[^ nil].
        
        (aBlock value:item) ifTrue:[
            ^ item.
        ].
    ].
    ^ nil

    "Created: / 13-08-2018 / 10:59:12 / Claus Gittinger"
!

identityIndexOf:aChild
    "return the index of aChild or 0 if not found. Compare using =="

    ^ self identityIndexOf:aChild startingAt:1
!

identityIndexOf:aChild startingAt:startIndex
    "return the index of aChild, starting search at startIndex.
     Compare using ==; return 0 if not found"

    |index|

    index := startIndex.

    self from:startIndex do:[:el|
        el == aChild ifTrue:[^ index ].
        index := index + 1.
    ].
    ^ 0
!

recursiveDetect:aOneArgBlock
    "recursive find the first child, for which evaluation 
     of the block returns true; if none, nil is returned.
     Warning: this only searches in already visible child elements
     i.e. any collapsed items are not searched."

    self recursiveDo:[:aChild|
        (aOneArgBlock value:aChild) ifTrue:[^ aChild]
    ].
    ^ nil
!

recursiveDetectLast:aBlock
    "recursive find the last child, for which evaluation of the block returns true; 
     if none does so, nil is returned.
     Warning: this only searches in already visible child elements
     i.e. any collapsed items are not searched."

    self recursiveReverseDo:[:aChild|
        (aBlock value:aChild) ifTrue:[^ aChild].
    ].
    ^ nil

    "Modified (comment): / 25-11-2016 / 08:41:42 / cg"
!

withAllDetect:aOneArgBlock
    "recursive find the first item including self, 
     for which evaluation of the block returns true; if none nil is returned.
     Warning: this only searches in already visible child elements
     i.e. any collapsed items are not searched."

    (aOneArgBlock value:self) ifTrue:[^ self].

    ^ self recursiveDetect:aOneArgBlock

    "Modified (comment): / 25-11-2016 / 08:41:53 / cg"
! !

!AbstractHierarchicalItem methodsFor:'sorting & reordering'!

recursiveSort:aSortBlock
    "recursive sort the children inplace using the 2-arg block sortBlock for comparison"

    children notEmptyOrNil ifTrue:[
        self synchronized:[
            "/ check again (asynchronous update was possible before)
            children notEmptyOrNil ifTrue:[
                self nonCriticalRecursiveSort:aSortBlock.
                self childrenOrderChanged.
            ]
        ]
    ].

    "Modified (format): / 28-07-2018 / 15:46:06 / Claus Gittinger"
!

sort:aSortBlock
    "sort the children inplace using the 2-arg block sortBlock for comparison"

    children notEmptyOrNil ifTrue: [
        self synchronized:[
            "/ check again (asynchronous update was possible before)
            children notEmptyOrNil ifTrue: [
                children := children sort:aSortBlock.
                self childrenOrderChanged.
            ]
        ].
    ]

    "Modified (format): / 28-07-2018 / 15:45:43 / Claus Gittinger"
! !

!AbstractHierarchicalItem class methodsFor:'documentation'!

version
    ^ '$Header$'
!

version_CVS
    ^ '$Header$'
! !

